################################################################################
#                                                                              #
# Copyright (c) 2020 VIA Technologies, Inc. All Rights Reserved.               #
#                                                                              #
# This PROPRIETARY SOFTWARE is the property of VIA Technologies, Inc.          #
# and may contain trade secrets and/or other confidential information of       #
# VIA Technologies, Inc. This file shall not be disclosed to any third         #
# party, in whole or in part, without prior written consent of VIA.            #
#                                                                              #
# THIS PROPRIETARY SOFTWARE AND ANY RELATED DOCUMENTATION ARE PROVIDED AS IS,  #
# WITH ALL FAULTS, AND WITHOUT WARRANTY OF ANY KIND EITHER EXPRESS OR IMPLIED, #
# AND VIA TECHNOLOGIES, INC. DISCLAIMS ALL EXPRESS OR IMPLIED                  #
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, QUIET       #
# ENJOYMENT OR NON-INFRINGEMENT.                                               #
#                                                                              #
################################################################################

import serial
import time
import sys
import json

FUNC_COLOR_DETECTION		= 1
FUNC_COLOR_CODE_DETECTION	= 2
FUNC_SHAPE_DETECTION		= 3
FUNC_SPHERE_DETECTION		= 4
FUNC_TEMPLATE_MATCHING		= 6
FUNC_KEYPOINTS				= 8
FUNC_NEURAL_NETWORK			= 9
FUNC_APRILTAG				= 10
FUNC_FACE_DETECTION			= 11
FUNC_TRAFFIC_SIGN_DETECTION	= 12
FUNC_HANDWRITTEN_DIGITS_DETECTION	= 13
FUNC_HANDWRITTEN_LETTERS_DETECTION	= 14
FUNC_REMOTE_COMPUTING		= 15
FUNC_LANES_DETECTION		= 16
FUNC_DIGITSOP_DETECTION		= 17
FUNC_SIMPLE_CLASSIFIER		= 18
FUNC_VOICE_COMMANDS			= 19

COLOR_RED		= 1
COLOR_YELLOW	= 2
COLOR_GREEN		= 3
COLOR_BLUE		= 4
COLOR_PURPLE	= 5
COLOR_BLACK		= 6

SHAPE_SPHERE	= 1
SHAPE_RECTANGLE	= 2
SHAPE_TRIANGLE	= 3
SHAPE_PENTAGON	= 4

ROI_EDGE		= 0
ROI_CLOUD		= 1
ROI_FACE		= 2
ROI_CENTRAL		= 3
ROI_BLOB		= 4
ROI_SALIENCY	= 5
ROI_BLOB_KALMAN	= 6

class Pixetto:
    def __init__(self):
        self.ser = None
        self.id  = 0
        self.num = 0
        self.objs_list = {}
        self.raw_data = ""

    def open(self, port):
        """
        Open a serial connection to Pixetto and initialize Pixetto. 
        Args:
            port: COM port of Pixetto.
        """
        self.ser = serial.Serial(port, 115200, timeout=3)

        #if self.ser.isOpen():
        #    print("OK")
        #else:
        #    print("failed")

        time.sleep(0.1)

        self.ser.write("{\"header\":\"STREAMON\"};".encode())
        self.ser.flushInput()

    def close(self):
        """
        Turn off Pixetto camera and close the serial connection.        
        """
        self.ser.write("{\"header\":\"STREAMOFF\"};".encode())
        self.ser.flushInput()
        self.ser.close()
        #print("closed!!")

    def is_detected(self):
        """
        Check if there is any object detected.
        Returns:
            True if any object is detected, False if none.
        """
        line = []

        count = self.ser.inWaiting()
        if count == 0:
            return False
        #print("count=", count)
        while count > 0:
            aa = self.ser.read(1)
            if aa == b';':
                break
            line.append(aa.decode("ascii"))
            count -= 1

        self.raw_data = "".join(x for x in line)
        #print("line=", self.raw_data)

        thing = json.loads(self.raw_data)

        if thing['header'] == 'DETECT' or thing['header'] == 'KEYPOINT':
            self.id  = thing['id']
            self.num = thing['num']
            if self.num > 0:
                if self.id == FUNC_VOICE_COMMANDS and thing['objects'][0]['type'] == 0:
                    self.num = 0
                    return False                
                self.objs_list = thing['objects']
                return True
            else:
                return False
        else:
            return False

    def get_data_list(self):
        """
        Parse the received data of detected objects.
        Returns: 
            id: function id
            num: number of detected objects
            objs_list: list of objects
        """
        return self.id, self.num, self.objs_list

    def get_raw_data(self):
        """
        Returns:
            Received raw data.        
        """
        return self.raw_data

    def send_command(self, cmdstr):
        print(cmdstr)
        self.ser.write(cmdstr.encode())
        self.ser.flushInput()
    
    def enable_color_detection(self, color=[], maxdisp=1, minsize=1):
        """
		Enable color detection.
        Args:
            color:   Set colors you want to detect. You can choose multiple colors at a time.
			         (COLOR_RED, COLOR_YELLOW, COLOR_GREEN, COLOR_BLUE, COLOR_PURPLE, COLOR_BLACK) 
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		    minsize: Set the minimal size as a threshold, to ignore the detect objects smaller than this size.
			         (range: 1~5  mapping to 10%,20%,30%,40%,50%) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_COLOR_DETECTION) + \
		             ',"maxdisp":' + str(maxdisp) + ',"minsize":' + str(minsize) + ',"color":"'
        for i in color:
        	config_str += str(i) + ','
        config_str += '"};'
        self.send_command(config_str)

    def enable_colorcode_detection(self, color=[], maxdisp=1, minsize=1):
        """
		Enable color code detection.
        Args:
            color:   Set colors you want to detect. In this function you must choose multiple colors at a time.
			         (COLOR_RED, COLOR_YELLOW, COLOR_GREEN, COLOR_BLUE, COLOR_PURPLE, COLOR_BLACK) 
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		    minsize: Set the minimal size as a threshold, to ignore the detect objects smaller than this size.
			         (range: 1~5  mapping to 10%,20%,30%,40%,50%) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_COLOR_CODE_DETECTION) + \
		             ',"maxdisp":' + str(maxdisp) + ',"minsize":' + str(minsize) + ',"color":"'
        for i in color:
        	config_str += str(i) + ','
        config_str += '"};'
        self.send_command(config_str)

    def enable_shape_detection(self, shape=[], maxdisp=1, minsize=1):
        """
		Enable shape detection.
        Args:
            shape:   Set shapes you want to detect. You can choose multiple shapes at a time.
			         (SHAPE_SPHERE, SHAPE_RECTANGLE, SHAPE_TRIANGLE, SHAPE_PENTAGON) 
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		    minsize: Set the minimal size as a threshold, to ignore the detect objects smaller than this size.
			         (range: 1~5  mapping to 10%,20%,30%,40%,50%) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_SHAPE_DETECTION) + \
		             ',"maxdisp":' + str(maxdisp) + ',"minsize":' + str(minsize) + ',"shape":"'
        for i in shape:
        	config_str += str(i) + ','
        config_str += '"};'
        self.send_command(config_str)

    def enable_sphere_detection(self, color=[], maxdisp=1, minsize=1):
        """
		Enable sphere detection.
        Args:
            color:   Set colors you want to detect. You can choose multiple colors at a time.
			         (COLOR_RED, COLOR_YELLOW, COLOR_GREEN, COLOR_BLUE, COLOR_PURPLE, COLOR_BLACK) 
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		    minsize: Set the minimal size as a threshold, to ignore the detect objects smaller than this size.
			         (range: 1~5  mapping to 10%,20%,30%,40%,50%) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_SPHERE_DETECTION) + \
		             ',"maxdisp":' + str(maxdisp) + ',"minsize":' + str(minsize) + ',"color":"'
        for i in color:
        	config_str += str(i) + ','
        config_str += '"};'
        self.send_command(config_str)

    def enable_template(self, maxdisp=1):
        """
		Enable template detection.
        Args:
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_TEMPLATE_MATCHING) + \
		             ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)
		 
    def enable_keypoint(self, hamming_dist=50, matched=5, maxdisp=1):
        """
		Enable template detection.
        Args:
            hamming_dist: Set the maximal hamming distance between the stored object and the detected object. 
            matched: Set the minimal number of keypoints matched
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_KEYPOINTS) + \
		             ',"hamming_dist":' + str(hamming_dist) + ',"matched":' + str(matched) + \
                     ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)
    
    def enable_neural_network(self, roi_type=ROI_CENTRAL, maxdisp=1):
        """
		Enable neural network.
        Args:
            roi_type: Choose a object detection algorithm
                      (ROI_EDGE, ROI_CLOUD, ROI_FACE, ROI_CENTRAL, ROI_BLOB, ROI_SALIENCY, ROI_BLOB_KALMAN)
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_NEURAL_NETWORK) + \
		             ',"roi_type":' + str(roi_type) + ',"maxdisp":' + str(maxdisp) + '};' 
        self.send_command(config_str)

    def enable_apriltag(self, maxdisp=1):
        """
		Enable apriltag detection.
        Args:
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_APRILTAG) + ',"maxdisp":' + str(maxdisp) + '};' 
        self.send_command(config_str)
        
    def enable_face_detection(self, maxdisp=1):
        """
		Enable face detection.
        Args:
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_FACE_DETECTION) + ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)
        
    def enable_traffic_sign_detection(self, maxdisp=1):
        """
		Enable traffic sign detection.
        Args:
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_TRAFFIC_SIGN_DETECTION) + ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)
        
    def enable_handwritten_digits_detection(self, maxdisp=1):
        """
		Enable handwritten digits detection.
        Args:
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_HANDWRITTEN_DIGITS_DETECTION) + ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)

    def enable_handwritten_letters_detection(self, maxdisp=1):
        """
		Enable handwritten letters detection.
        Args:
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_HANDWRITTEN_LETTERS_DETECTION) + ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)

    def enable_remote_computing(self, host="mls.pixetto.ai", target="/PixServerYolov3/", port=80, maxdisp=1):
        """
		Enable remote computing.
        Args:
            host: host name
            target: service name
            port: port number
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_REMOTE_COMPUTING) + \
		             ',"host":"' + host + '","target":"' + target + '","port":"' + str(port) + \
		             '","maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)

    def enable_lanes_detection(self, black_lanes=False):
        """
		Enable lanes detection.
        Args:
            black_lanes: True for black lanes, and False for white lanes
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_LANES_DETECTION) + ',"black_lanes":'
        if black_lanes == True:
            config_str += 'true'
        else:
            config_str += 'false'
        config_str += '};'
        self.send_command(config_str)
        
    def enable_digits_operation_detection(self, maxdisp=1):
        """
		Enable digits operation detection.
        Args:
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_DIGITSOP_DETECTION) + ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)

    def enable_simple_classifier(self, roi_type=ROI_CENTRAL, maxdisp=1):
        """
		Enable simple classifier.
        Args:
            roi_type: Choose a object detection algorithm
                      (ROI_EDGE, ROI_CLOUD, ROI_FACE, ROI_CENTRAL, ROI_BLOB, ROI_SALIENCY, ROI_BLOB_KALMAN)
		    maxdisp: Set the maximal number of objects you want to detect (range : 1~5) 
		""" 
        config_str = '{"header":"FUNC","id":' + str(FUNC_SIMPLE_CLASSIFIER) + \
		             ',"roi_type":' + str(roi_type) + ',"maxdisp":' + str(maxdisp) + '};'
        self.send_command(config_str)
        
    def enable_voice_commands(self):
        """
		Enable voice commands.
		"""
        config_str = '{"header":"FUNC","id":' + str(FUNC_VOICE_COMMANDS) + '};' 
        self.send_command(config_str)
